/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.strawberry;

import java.text.ParseException;

import java.io.IOException;
import java.io.BufferedReader;
import java.io.InputStreamReader;

import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;

import org.ucam.srcf.assassins.AssassinActivation;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.AssassinStateViewer;
import org.ucam.srcf.assassins.ConfigProperty;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.Plugin;
import org.ucam.srcf.assassins.StateChange;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.StateBuilderException;
import org.ucam.srcf.assassins.xml.Config;
import org.ucam.srcf.assassins.Assassin;
import org.ucam.srcf.assassins.PluginCommand;
import org.ucam.srcf.assassins.TextQuestion;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.Death;
import org.ucam.srcf.assassins.Resurrection;
import org.ucam.srcf.assassins.PluginStateChange;
import org.ucam.srcf.assassins.Event;
import org.ucam.srcf.assassins.PlayerViewer;
import org.ucam.srcf.assassins.Tools;

import org.ucam.srcf.assassins.exceptions.ConfigFileException;

/**
 * Interface for java classes which act as plug-in game modifiers.
 */
public class Strawberry extends Plugin
{
   public static String getDescription() { return "Scoring plugin"; };
	public static int defaultStartingBaseScore;
	/** Start time of game, in millis*/
	long startTime;
	
	Scorer AbaseScore;
	Scorer VbaseScore;
	Scorer AScoreMod;
	Scorer VScoreMod;
	
   /**
    * Sets the canonical string representing the Plugin
    * @param s The string representing the Plugin
    */
   public void setPluginName(String s) {
		super.setPluginName(s);
		
		if (Debug.debug) Debug.print(this, "setting plugin name");

		try {
		   defaultStartingBaseScore = Config.getConfigInt("Strawberry", "default-starting-score");
			AbaseScore=new Scorer(Config.getConfig("Strawberry", "AbaseScoreEqn"),0);
			VbaseScore=new Scorer(Config.getConfig("Strawberry", "VbaseScoreEqn"),0);
			AScoreMod=new Scorer(Config.getConfig("Strawberry", "AScoreModEqn"),0);
			VScoreMod=new Scorer(Config.getConfig("Strawberry", "VScoreModEqn"),0);
			startTime = Tools.unFormatTime(Config.getConfig("vanilla", "game-start-date"));
		}
		catch (ConfigFileException cfex) {
			if(Debug.debug) Debug.print(this, "Dodgy config file - "+cfex);
			System.out.println("Config file error:");
			cfex.printStackTrace();
		}
		catch (ParseException pex) {
			if(Debug.debug) Debug.print(this, "Parse exception when formattign time - "+pex);
			System.out.println("ParseError while reading config:");
			pex.printStackTrace();
		}
	}
   
   /**
    * Called whenever the plugin is required to update or sanity-check the
    * current gamestate.
    */
   protected Vector updateState(Map state, Vector qs, EventViewer et) 
		throws AssassinSQLException, ParseException 
	{ 
		if(Debug.debug) Debug.print(this, "Updating State by rescoring all assassins");

      /* Run the entire game and validate the scores */
      Set allAss = new HashSet(state.keySet());
      scoreFromSetFromTime(allAss, startTime);

		if(Debug.debug) Debug.print(this, "Done Updating State");
		return null; 
   }

   /**
    * Scores the game from a given time, only inspecting assassins in a given
    * set.
    */
   private void scoreFromSetFromTime(Set interested, long fromTime) throws AssassinSQLException {
		/* We proceed to go through the system and check the scores 
		 * assigned at each "death" event. This officially sucks, as we have
		 * to build our own state for aliveness
		 */
		/* I have been assured that noone would be silly enough to think
		 * that there should eb intermediate scores worked out during an
		 * event, and that noone will put a chain of deaths
		 * A->B->C and expect A to get some of B's points for killing
		 * C from this event.
		 */

		/* Set up a var tracker to track stuff */
		VarTracker vt = new VarTracker(interested, getPluginName());

		/* Get all events in chronological order */
		List events = Event.getEventViewersInTime(fromTime, vanilla.getNowTime());

      /* Read in starting scores if we're not at the beginning */
      if(startTime != fromTime) {
         if(Debug.debug) Debug.print(this, "Reading in starting scores for assassins");
         Iterator i = interested.iterator();
         while(i.hasNext()) {
            vt.readValuesForAssassinAtTime((AssassinViewer) i.next(), fromTime);
         }
      }
      
		Iterator i = events.iterator();
		if(Debug.debug) Debug.print(this, "# events in time ("+fromTime+","+vanilla.getNowTime()+") = "+events.size());
		while(i.hasNext()) {
			EventViewer e = (EventViewer) i.next();
			if(!e.isActive())
				continue;
			if(Debug.debug) Debug.print(this, "Processing event \""+e.getHeadline()+"\" at time "+Tools.prettyDateAndTime(e.getTime()));
         if(!eventContainsParticipantsFromSet(e, interested)) {
            if(Debug.debug) Debug.print(this, "Doesn't contain anything we're interested in!");
            continue;
         }
			List sclist = e.getStateChanges();
			Iterator j = sclist.iterator();

         /* Flag for assassinActivation events */
         boolean assAct = false;

			/*Set up a map from assassins to *modified* scores*/
			Map scores = vt.refreshScores(e);
			
			/* List of score StateChanges so we can check if they need updating at 
			 * the end of processing an event
			 */
			List scoreChanges = new ArrayList();

			/*Now iterate through the statechanges*/
			while(j.hasNext()) {
				StateChange sc = (StateChange) j.next();
            if(!doesSCInvolveSet(sc, interested)) {
               if(Debug.debug) Debug.print(this, "We don't care about "+
                     "this one: " + sc);
               continue;
            }
				AssassinViewer a = sc.getAssassinViewer();
				String pname = sc.getPluginName();
				String field = sc.getFieldName();
				Object value = sc.getValue();
            if(sc instanceof AssassinActivation) {
               assAct = true;
               continue;
            }
				if(sc instanceof Death) {
					if(Debug.debug) Debug.print(this, "SC is a death");
					/* Make certain not already dead, assume alive if return is ""*/
					if(vt.getProperty(a,"alive").equals("false")) {
                  if(Debug.debug) Debug.print(this, "...but vicitim is already dead");
						continue;
               }
					vt.setProperty(a,"alive","false");
               /* Check if people were in the interested set and add 
                * them as appropriate, loading info */
               if(!interested.contains(a)) {
                  interested.add(a);
                  vt.readValuesForAssassinAtTime(a, e.getTime());
               }
               if(!interested.contains(((Death) sc).getKiller())) {
                  interested.add((((Death) sc).getKiller()));
                  vt.readValuesForAssassinAtTime(((Death) sc).getKiller(), 
                        e.getTime());
               }
					/* Score it and notify the VarTracker. */
					scoreDeath((Death) sc, vt, scores);
					vt.trackDeath((Death) sc);
					continue;
				} 
				if(sc instanceof Resurrection) {
					if(Debug.debug) Debug.print(this, "SC is a resurrection");
					vt.setProperty(a,"alive","true");
					continue;
				} 
				if(!pname.equals(this.getPluginName()))
					continue;
				/* Otherwise we have a sc applicable to this plugin, */
				if(Debug.debug) Debug.print(this, "SC is something else Strawberry related: "+field);

				/* If its a score change we deal with it later */
				if(	field.equals("BaseScore") ||
						field.equals("PostGameMod") ||
						field.equals("FinalScore")
						) {
					scoreChanges.add(sc);
					continue;
				}

				/* If its an innocent death we deal with it */
				if(field.equals("Innocents")) {
					int numKilled = Integer.parseInt((String) value);
					for(int innocent=0; innocent<numKilled; innocent++) {
						scoreInnocentDeath(a, vt, scores);
						vt.trackInnocent(a);
					}
				}

				/*If its a bonus we add it to the scores*/
				if(field.equals("BaseBonus")) {
					if(Debug.debug) Debug.print(this, "Applying BaseBonus ...");
					AssassinScore as=((AssassinScore) scores.get(a));
					if(as==null) 
						scores.put(a, as=new AssassinScore(a, this.getPluginName(), vt));
					as.addBase(Integer.parseInt((String) value));
					if(Debug.debug) Debug.print(this, "...done");
					continue;
				}
				if(field.equals("ModBonus")) {
					if(Debug.debug) Debug.print(this, "Applying ModBonus ...");
					AssassinScore as=((AssassinScore) scores.get(a));
					if(as==null)
						scores.put(a, as=new AssassinScore(a, this.getPluginName(), vt));
					as.addPost(Integer.parseInt((String) value));
					if(Debug.debug) Debug.print(this, "...done");
					continue;
				}
				
				/* otherwise we just apply it */
				vt.setProperty(a, field, (String) value);
				if(Debug.debug) Debug.print(this, "Applying unknown Strawberry State Change (" + field + ", " + value + ")");
			} /*End of states loop */
			if(Debug.debug) Debug.print(this, "Done with states loop.");

			/* 
          * Now we go through and check if the scores are right or not
          * by going first going through preexisting ones and chekcing 
          * them, removing them if we have no record of them!
          */
			j = scoreChanges.iterator();
			while(j.hasNext()) {
				StateChange sc = (StateChange) j.next();
				AssassinViewer a = sc.getAssassinViewer();
				String field = sc.getFieldName();
				int value = Integer.parseInt((String) sc.getValue());
				if(scores.get(a) == null) {
               if(!assAct) {
                  /* Its wrong! remove it! */
                  vanilla.removeStateChange(e, sc);
               } else {
                  /* Its part of an assassin activation event */
                  vt.setProperty(a, field, String.valueOf(value));
               }
					continue;
				}
				AssassinScore cur = (AssassinScore) scores.get(a);
				if(field.equals("BaseScore")) {
					if(cur.getBaseScore() != value)
						vanilla.removeStateChange(e, sc);
					else
						cur.baseClean();
					continue;
				}
				if(field.equals("PostGameMod")) {
					if(cur.getPostGameMod() != value)
						vanilla.removeStateChange(e, sc);
					else
						cur.postClean();
					continue;
				}
				if(field.equals("FinalScore")) {
					if(cur.getFinalScore() != value)
						vanilla.removeStateChange(e, sc);
					else
						cur.finalClean();
					continue;
				}
			}
			if(Debug.debug) Debug.print(this, "Done with initial score storing loop.");
			/* Now we add any dirty score changes to the event */
			j = scores.keySet().iterator();
			while(j.hasNext()) {
				AssassinScore cur=(AssassinScore) scores.get(j.next());
				if(cur==null)
					continue;
				AssassinViewer a = cur.getAssassinViewer();
				/*write out if they need it*/
				if(cur.isBaseDirty())
					vanilla.addStateChange(e, cur.getBaseSC(this));
				if(cur.isPostDirty())
					vanilla.addStateChange(e, cur.getPostSC(this));
				if(cur.isFinalDirty())
					vanilla.addStateChange(e, cur.getFinalSC(this));
				/*and update our internal state*/
				vt.setIntProperty(a, "BaseScore", cur.getBaseScore());
				vt.setIntProperty(a, "PostGameMod", cur.getPostGameMod());
				vt.setIntProperty(a, "FinalScore", cur.getFinalScore());
			}
			if(Debug.debug) Debug.print(this, "Done with final score storing loop.");
				
		} /*end of events loop */
	}

	private void scoreInnocentDeath(AssassinViewer a, VarTracker vt, Map scores) {
		int newAbaseScore = (int) Math.round(AbaseScore.execute(a, null, vt));
		int newApostGameMod = (int) Math.round(AScoreMod.execute(a, null, vt));
		
		AssassinScore aScore = (AssassinScore) scores.get(a);

		if(aScore==null)
			scores.put(a, aScore=new AssassinScore(a, this.getPluginName(), vt));
		
		aScore.setBaseScore(newAbaseScore);
		aScore.setPostGameMod(newApostGameMod);
	}
	
	private void scoreDeath(Death myDeath, VarTracker vt, Map scores) {
      /* calculate the new scores */
      try {
			if(Debug.debug) Debug.print(this, "scoring death: "+myDeath.getKiller().getPlayerViewer().getName()+"->"+myDeath.getVictim().getPlayerViewer().getName());
         int newAbaseScore = (int) Math.round(AbaseScore.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));
         int newVbaseScore = (int) Math.round(VbaseScore.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));
         int newApostGameMod = (int) Math.round(AScoreMod.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));
         int newVpostGameMod = (int) Math.round(VScoreMod.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));

			if(Debug.debug) {
				Debug.print(this, "\tAbaseScore="+newAbaseScore);
				Debug.print(this, "\tApostGameMod="+newApostGameMod);
				Debug.print(this, "\tVbaseScore="+newVbaseScore);
				Debug.print(this, "\tApostGameMod="+newVpostGameMod);
			}

         /* get the current scores if they exist, otherwise make them */
         AssassinScore aScore = (AssassinScore) scores.get(myDeath.getKiller());
         AssassinScore vScore = (AssassinScore) scores.get(myDeath.getVictim());
         if (aScore == null)
               scores.put(myDeath.getKiller(), aScore = new AssassinScore(
                     myDeath.getKiller(), this.getPluginName(), vt));
         if (vScore == null)
               scores.put(myDeath.getVictim(), vScore = new AssassinScore(
                     myDeath.getVictim(), this.getPluginName(), vt));
         /* now set the scores correctly */
         aScore.setBaseScore(newAbaseScore);
         aScore.setPostGameMod(newApostGameMod);

         vScore.setBaseScore(newVbaseScore);
         vScore.setPostGameMod(newVpostGameMod);
      } catch (AssassinSQLException asqle) {
			if(Debug.debug) Debug.print(this, "Exception while scoring deaths: "+asqle);
         return;
      }
		if(Debug.debug) Debug.print(this, "Done scoring death.");
   }

   /**
    * Called when an event is being added to get any queries the plugin
    * has for the user. The plugin should check all the event details and if it
    * needs more information it should ask relevent questions. The 
    * returnQuestions function is passed the same Vector with the answers 
    * filled out.
    * @param e The Event which is being added
    * @return A Vector of Questions
    * @see returnQuestions
    */
   protected Vector participantAdded(EventViewer e, AssassinViewer a, Map state, Vector qs) 
		throws AssassinSQLException 
   {
      if (null == qs) {
         qs = new Vector();

         PlayerViewer p = a.getPlayerViewer();
         qs.add(new TextQuestion("Give " + p.getName() + 
                  " a transfferable bonus of:", "0"));
         qs.add(new TextQuestion("Give " + p.getName() + 
                  " a non-transfferable bonus of:", "0"));
         qs.add(new TextQuestion("Number of innocents " + p.getName() + 
                  " killed:", "0"));
         return qs;
      } else {
         /* set of assassins which potentially have a change in score */
         Set startSet = new HashSet();

         PlayerViewer p = a.getPlayerViewer();
         String baseScoreBonus = ((TextQuestion) qs.get(0)).getAnswer();
         String postGameBonus = ((TextQuestion) qs.get(1)).getAnswer();
         String innocents = ((TextQuestion) qs.get(2)).getAnswer();
         if(!baseScoreBonus.equals("0")) {
            vanilla.addStateChange(e, new PluginStateChange(a, this, "BaseBonus", baseScoreBonus));
            startSet.add(a);
         }
         if(!postGameBonus.equals("0")) {
            vanilla.addStateChange(e, new PluginStateChange(a, this, "ModBonus", postGameBonus));
            startSet.add(a);
         }
         if(!innocents.equals("0")) {
            vanilla.addStateChange(e, new PluginStateChange(a, this, "Innoccents", innocents));
            startSet.add(a);
         }
         /* Now update the scores */
         scoreFromSetFromTime(startSet, e.getTime());
         return null;
      }
   }

   protected Vector stateChangeAdded(StateChange sc, Map state, Vector qs) throws AssassinSQLException
   {
      if (null == qs) {
         HashSet startSet = new HashSet();
         if(sc instanceof Death) {
            startSet.add(((Death) sc).getKiller());
            startSet.add(((Death) sc).getVictim());
         }
         /* Now update the scores */
         scoreFromSetFromTime(startSet, sc.getEventViewer().getTime());
         return new Vector();
      } else {
         return null;
      }
   }

   /**
    * Called when an assassin is being added to get any queries the plugin
    * has for the user. The plugin should check all the assassin details and if it
    * needs more information it should ask relevent questions. The 
    * returnQuestions function is passed the same Vector with the answers 
    * filled out.
    * @param a The Assassin which is being added
    * @return A Vector of Questions
    * @see returnQuestions
    */
   protected Vector assassinAdded(AssassinViewer a, EventViewer e, Map state, Vector qs) 
      throws AssassinSQLException 
   {
      if (null == qs) {
         qs=new Vector();

         qs.add(new TextQuestion("Starting Base Score", ""+defaultStartingBaseScore));
         qs.add(new TextQuestion("Starting Non-transferable Score Modifier", "0"));

         return qs;
      } else {

         /* We expect the following questions:
          * 0	-	Starting Base Score			(BaseScore)
          * 1	-	Starting Score Modifier		(PostGameMod)
          */

         Debug.print(this, "Returned "+qs.size()+" questions");
         int baseScore = Integer.parseInt(((TextQuestion) qs.get(0)).getAnswer());
         int postGameMod = Integer.parseInt(((TextQuestion) qs.get(1)).getAnswer());

         vanilla.addStateChange(e, new PluginStateChange(a, this, "BaseScore", String.valueOf(baseScore)));
         vanilla.addStateChange(e, new PluginStateChange(a, this, "PostGameMod", String.valueOf(postGameMod)));
         vanilla.addStateChange(e, new PluginStateChange(a, this, "FinalScore", String.valueOf(baseScore+postGameMod)));
         Debug.print(this, "Done with Qs");
         return null;
      }
   }

   /**
    * Returns a Map of new commands to add to the CLI
    *
    * @return A Map of String =&gt; Command
    */
   public Map getCommands() { 
		Map m = new HashMap();
		final Strawberry st = this;
		
		m.put("score-assassin",
				new PluginCommand() {
					public void pluginCommand(Map params) throws AssassinSQLException {
						st.scoreAssassin(params);
					}
				}
		);
		m.put("static-score",
				new PluginCommand() {
					public void pluginCommand(Map params) throws AssassinSQLException {
						st.staticScore(params);
					}
				}
		);
		
		return m; 
	}

   /**
    * Returns a Map of new commands to add to the CGI
    *
    * @return A Map of String =&gt; CGICommand
    */
   public Map getCGICommands() { 
		return new HashMap(); 
	}

	private AssassinViewer selectAssassin(Map state) 
		throws IOException, AssassinSQLException {
		List names=new ArrayList();

		Iterator i=state.keySet().iterator();
		while(i.hasNext())
			names.add(((AssassinViewer) i.next()).getPlayerViewer().getName());
		
		int index = askOption("Select Assassin:", names, 0) - 1;
		String name = (String) names.get(index);
		
		i=state.keySet().iterator();
		AssassinViewer current=null;
		while(i.hasNext())
			if((current=((AssassinViewer) i.next())).getPlayerViewer().getName().equals(name))
				return current;
					
		return null;
	}

	private int askOption(String prompt, List options, int def) 
		throws IOException {
		boolean notDone=true;
		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
		int ans=0;
		
		while(notDone) {
			System.out.println(prompt);
			Iterator i=options.iterator();
			int j=0;
			while(i.hasNext())
				System.out.println((j++) + ". " + ((String) i.next())); 
			if(def!=0)
				System.out.print("Enter a number ["+def+"]:");
			else
				System.out.print("Enter a number:");
			String selection=in.readLine();
			if(selection.equals(""))
				return def;
			try {
				if(	(ans=Integer.parseInt(selection)) > 0 &&
						Integer.parseInt(selection) < options.size()		)
					notDone=false;
				else
					notDone=true;
			} catch (NumberFormatException ex) {
				notDone=true;
			}
		}
		return ans;
	}

   /**
    * Does the given StateChagne have anything to do with anyone in the set
    * involving?
    * @return true if it does, false otherwise.
    */
   private boolean doesSCInvolveSet(StateChange sc, Set involving) throws AssassinSQLException {
      if(sc instanceof Death && involving.contains(((Death) sc).getKiller()))
         return true;
      if(involving.contains(sc.getAssassinViewer()))
         return true;
      return false;
   }

   /**
    * Does the given event have any particiapnts from the given Set?
    */
   private boolean eventContainsParticipantsFromSet(EventViewer e, Set s) throws AssassinSQLException {
      Iterator i = e.getStateChanges().iterator();
      while(i.hasNext()) {
         StateChange sc = (StateChange) i.next();
         if(sc instanceof Death)
            if(s.contains(((Death) sc).getKiller()))
               return true;
         if(s.contains(sc.getAssassinViewer()))
            return true;
      }
      return false;
   }

	public void scoreAssassin(Map params) 
		throws AssassinSQLException 
	{
      // TODO: should be able to only build state for one assassin here?
      Map state;
	  try {
		  state = vanilla.getState();
	  } catch (StateBuilderException e) {
		  System.out.println("Failed to get Vanilla State, exiting.");
		  return;
	  }
		try {
			updateState(state, null, null);
		}
		catch (ParseException pe) {
			System.out.println("Failed to build Stawberry state, exiting.");
			return;
		}
		
		AssassinViewer assassin;
		try {
			assassin = selectAssassin(state);
		} catch (IOException IOe) {
			System.out.println("Error while readin stdin, terminating.");
			return;
		}
		AssassinStateViewer aState = (AssassinStateViewer) state.get(assassin);
		
		System.out.println("Scoring assassin " + aState.getPseudonyms().get(aState.getPseudonyms().size()));
		
		int baseScore = aState.getIntProperty(this.getPluginName(), "BaseScore");
		int postGameMod = aState.getIntProperty(this.getPluginName(), "PostGameMod");
		int finalScore = aState.getIntProperty(this.getPluginName(), "FinalScore");
		
		System.out.println("Final Score = Base Score + Post Game Modifier");
		System.out.println(finalScore + " = " + baseScore + " + " + postGameMod);
	}
   
   public static Vector getConfigProperties() {
      Vector v = new Vector();
      v.add ( new ConfigProperty ("default-starting-score", "", ""));
      v.add ( new ConfigProperty ("AbaseScoreEqn", "", ""));
      v.add ( new ConfigProperty ("VbaseScoreEqn", "", ""));
      v.add ( new ConfigProperty ("AScoreModEqn", "", ""));
      v.add ( new ConfigProperty ("VScoreModEqn", "", ""));
	  return v;
   }

   public void staticScore(Map params) throws AssassinSQLException {
      /* Get a list of all assassins */
      List assassins=Assassin.getAllAssassinViewers();

      /* Create our scoring object */
      StaticScoringAlgorithm statScorer=new KirikaStatic();
      statScorer.init(assassins.size());

      /* Build state */
      Set dead=new HashSet();
      if(Debug.debug) Debug.print(this, "Building state");
      List events = Event.getEventViewersInTime(startTime, vanilla.getNowTime());
      Iterator i = events.iterator();
      if(Debug.debug) Debug.print(Strawberry.class, "# events in time ("+startTime+","+vanilla.getNowTime()+") = "+events.size());
      while(i.hasNext()) { 
         EventViewer e = (EventViewer) i.next();
         if(!e.isActive())
            continue;
         if(Debug.debug) Debug.print(Strawberry.class, "Processing event \""+e.getHeadline()+"\" at time "+Tools.prettyDateAndTime(e.getTime()));
         List sclist = e.getStateChanges();
         Iterator j = sclist.iterator();
         while(j.hasNext()) {
            StateChange sc = (StateChange) j.next();
            if(sc instanceof Death && !dead.contains(((Death) sc).getVictim())) {
               statScorer.addDeath((Death) sc);
               dead.add(((Death) sc).getVictim());
            }
            if(sc instanceof Resurrection) {
               dead.remove(sc.getAssassinViewer());
            }
         }
      }

      /* Score */
      if(Debug.debug) Debug.print(this, "Scoring");
      Map scores=statScorer.score();

      /* Print the answers */
      i=assassins.iterator();
      while(i.hasNext()) {
         AssassinViewer current=(AssassinViewer) i.next();
         Double score = (Double) scores.get(current);
         if(score==null)
            score=new Double(1);
         System.out.println(current.getPlayerViewer().getName()+" : "+score.doubleValue());
      }
   }
}


